// dma.c
// DUAL3006 DMA-based application for 5510 DSK
// Educational DSP, LLC - 2006

// DMA configuration and handlers, SWI for overrun handling

#include "dsk5510_dual3006cfg.h"
#include "dsk5510.h"
#include <csl.h>
#include <csl_irq.h>
#include <csl_dma.h>
#include "swi_process.h"


#pragma DATA_SECTION(receive_buffer, "inBuffer")
short receive_buffer[BUFFER_DEPTH][BUFFER_LENGTH * NUM_CHANNELS];
#pragma DATA_SECTION(transmit_buffer, "outBuffer")
short transmit_buffer[BUFFER_DEPTH][BUFFER_LENGTH * NUM_CHANNELS];

// assigned initial index values that are used in init_DMA()
short dma_receive_buffer_index = 0;
short dma_transmit_buffer_index = 0;

Uint16 receive_buffer_processed = 1;
Uint16 transmit_buffer_filled = 1;
Uint16 receive_buffer_to_process_index = 0;
Uint16 transmit_buffer_to_fill_index = 0;

void init_DMA()
{
	Uint16 eventIdRcv, eventIdXmt;
	volatile int i;
	Uint32 x;

    // set indices and lengths for receive channel
    DMA_RSETH(hDmaRcv, DMACEI, (-MCBSP1_0_OFFSET)-3);
    DMA_RSETH(hDmaRcv, DMACFI, (MCBSP1_0_OFFSET)-3);
    DMA_RSETH(hDmaRcv, DMACFN, BUFFER_LENGTH);
	// set initial buffer address
	x = (Uint32)(&receive_buffer[dma_receive_buffer_index][0]);
	x <<= 1; // need byte_address for DMA
    DMA_RSETH(hDmaRcv, DMACDSAL, x & 0xffff);
    DMA_RSETH(hDmaRcv, DMACDSAU, (x >> 16) & 0xffff);

    // Set indices and lengths for transmit channel
    DMA_RSETH(hDmaXmt, DMACEI, (-MCBSP1_0_OFFSET)-3);
    DMA_RSETH(hDmaXmt, DMACFI, (MCBSP1_0_OFFSET)-3);
    DMA_RSETH(hDmaXmt, DMACFN, BUFFER_LENGTH);
	// set initial buffer address
	x = (Uint32)(&transmit_buffer[dma_transmit_buffer_index][0]);
	x <<= 1; // need byte_address for DMA
    DMA_RSETH(hDmaXmt, DMACSSAL, x & 0xffff);
    DMA_RSETH(hDmaXmt, DMACSSAU, (x >> 16) & 0xffff);


    // Clear the DMA status registers to receive new interrupts
    i = DMA_RGETH(hDmaRcv, DMACSR);
    i = DMA_RGETH(hDmaXmt, DMACSR);

    // Get Event ID associated with DMA channel interrupt.  Event IDs are a
    // CSL abstraction that lets code describe a logical event that gets
    // mapped to a real physical event at run time.  This helps to improve
    // code portability.
    eventIdRcv = DMA_getEventId(hDmaRcv);
    eventIdXmt = DMA_getEventId(hDmaXmt);

    // Clear any pending receive channel interrupts (IFR)
    IRQ_clear(eventIdRcv);
    IRQ_clear(eventIdXmt);

    // Enable receive DMA interrupt (IMR)
    IRQ_enable(eventIdRcv);
    IRQ_enable(eventIdXmt);

    // Make sure global interrupts are enabled
    IRQ_globalEnable();

	// start DMA channels
    DMA_start(hDmaRcv);
    DMA_start(hDmaXmt);
}

void HWI_DMA0_Transmit()
{
	Uint32 x;

	if(!transmit_buffer_filled) {	// check for overrun
		SWI_or(&SWI_Overrun, TRANSMIT_OVERRUN);
	}
	// calculate buffer to use on next reload of DMA
	// DMA has already auto-reloaded with the address
	// programmed in the last HWI invocation
	// update ping-pong buffer index
	++dma_transmit_buffer_index;
	dma_transmit_buffer_index &= 1;
	// need to fill that buffer now while other buffer is being used
	transmit_buffer_to_fill_index = dma_transmit_buffer_index;
	transmit_buffer_filled = 0;
	// set address for next reload
	x = (Uint32)(transmit_buffer[dma_transmit_buffer_index]);
	x <<= 1; // need byte_address for DMA
    DMA_RSETH(hDmaXmt, DMACSSAL, x & 0xffff);
    DMA_RSETH(hDmaXmt, DMACSSAU, (x >> 16) & 0xffff);

    // Read the DMA status register to clear it so new interrupts will be seen
    DMA_RGETH(hDmaXmt, DMACSR);

	// post SWI, will use transmit_buffer_to_fill_index
	SWI_or(&SWI_Process, DMA_TRANSMIT_DONE);
}

void HWI_DMA1_Receive()
{
	Uint32 x;

	if(!receive_buffer_processed) {	// check for overrun
		SWI_or(&SWI_Overrun, RECEIVE_OVERRUN);
	}
	// calculate buffer to use on next reload of DMA
	// DMA has already auto-reloaded with the address
	// programmed in the last HWI invocation
	// update ping-pong buffer index
	++dma_receive_buffer_index;
	dma_receive_buffer_index &= 1;
	// need to process that buffer now while other buffer is being used
	receive_buffer_to_process_index = dma_receive_buffer_index;	
	receive_buffer_processed = 0;
	// set address for next reload
	x = (Uint32)(receive_buffer[dma_receive_buffer_index]);
	x <<= 1; // need byte_address for DMA
    DMA_RSETH(hDmaRcv, DMACDSAL, x & 0xffff);
    DMA_RSETH(hDmaRcv, DMACDSAU, (x >> 16) & 0xffff);

    // Read the DMA status register to clear it so new interrupts will be seen
    DMA_RGETH(hDmaRcv, DMACSR);

	// post SWI, will use receive_buffer_to_process_index
	SWI_or(&SWI_Process, DMA_RECEIVE_DONE);
}

void SWI_OverrunHandler()
{
	unsigned int i;

	i = SWI_getmbox();
	if(i & RECEIVE_OVERRUN)
		LOG_printf(&LOG_errors, "DMA Receive overrun error - DMA halted");
	if(i & TRANSMIT_OVERRUN)
		LOG_printf(&LOG_errors, "DMA Transmit overrun error - DMA halted");

	// disable DMA
    DMA_stop(hDmaRcv);
    DMA_stop(hDmaXmt);
}
